#pragma once

#include <list>
#include <memory>

#include <amdis/Boundary.hpp>
#include <amdis/LocalAssemblerBase.hpp>
#include <amdis/utility/TreeData.hpp>

namespace AMDiS
{
  namespace tag
  {
    template <class E> struct element_operator { using type = E; };
    template <class I> struct intersection_operator { using type = I; };
    template <class I> struct boundary_operator { using type = I; BoundaryType id; };
  }

  template <class GridView>
  class OperatorLists
  {
    using Element = typename GridView::template Codim<0>::Entity;
    using Intersection = typename GridView::Intersection;

    template <class OperatorType>
    struct DataElement
    {
      std::shared_ptr<OperatorType> op;
      BoundaryType b = {0};
    };

    /// Lists of \ref DataElement on the Element, BoundaryIntersction, and InteriorIntersections
    template <class... Nodes>
    struct Data
    {
      /// The type of local operators associated with grid elements
      using ElementOperator = LocalAssemblerBase<Element, Nodes...>;
      /// The type of local operators associated with grid intersections
      using IntersectionOperator = LocalAssemblerBase<Intersection, Nodes...>;

      /// Return whether any operators are stored on the node
      bool empty() const
      {
        return element.empty() && boundary.empty() && intersection.empty();
      }

      /// Test whether to assemble on the node
      bool doAssemble() const
      {
        return !assembled || changing;
      }

      /// Bind all operators to the grid element and geometry
      template <class Geo>
      void bind(Element const& elem, Geo const& geo)
      {
        for (auto& scaled : element) scaled.op->bind(elem,geo);
        for (auto& scaled : boundary) scaled.op->bind(elem,geo);
        for (auto& scaled : intersection) scaled.op->bind(elem,geo);
      }

      /// Unbind all operators from the element
      void unbind()
      {
        for (auto& scaled : element) scaled.op->unbind();
        for (auto& scaled : boundary) scaled.op->unbind();
        for (auto& scaled : intersection) scaled.op->unbind();
      }


      template <class... Args>
      void push(tag::element_operator<Element>, Args&&... args)
      {
        element.push_back({std::forward<Args>(args)...});
      }

      template <class... Args>
      void push(tag::boundary_operator<Intersection> b, Args&&... args)
      {
        boundary.push_back({std::forward<Args>(args)..., b.id});
      }

      template <class... Args>
      void push(tag::intersection_operator<Intersection>, Args&&... args)
      {
        intersection.push_back({std::forward<Args>(args)...});
      }


      /// List of operators to be assembled on grid elements
      std::list<DataElement<ElementOperator>> element;
      /// List of operators to be assembled on boundary intersections
      std::list<DataElement<IntersectionOperator>> boundary;
      /// List of operators to be assembled on interior intersections
      std::list<DataElement<IntersectionOperator>> intersection;

      /// if false, do reassemble of all operators
      bool assembled = false;

      /// if true, or \ref assembled false, do reassemble of all operators
      bool changing = false;
    };

  public:

    /// List of operators associated with matrix blocks (RowNode, ColNode)
    template <class RowNode, class ColNode>
    using MatData = Data<RowNode, ColNode>;

    /// List of operators associated with vector blocks [Node]
    template <class Node>
    using VecData = Data<Node>;
  };


  template <class RowBasis, class ColBasis>
  using MatrixOperators = MatrixData<RowBasis, ColBasis, OperatorLists<typename RowBasis::GridView>::template MatData>;

  template <class GlobalBasis>
  using VectorOperators = VectorData<GlobalBasis, OperatorLists<typename GlobalBasis::GridView>::template VecData>;

} // end namespace AMDiS
