#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/LocalBasisCache.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_trial {};
  }


  /// zero-order operator \f$ \langle\psi, c\,\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::test_trial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_trial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test_trial, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix);
      else
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix);
    }

  protected:
    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;
      LocalBasisCache<RowLocalBasisType> rowCache(rowLocalFE.localBasis());
      LocalBasisCache<ColLocalBasisType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQp(context,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQp(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowSize; ++i) {
          const auto local_i = rowNode.localIndex(i);
          const auto value = factor * rowShapeValues[i];

          for (std::size_t j = 0; j < colSize; ++j) {
            const auto local_j = colNode.localIndex(j);
            elementMatrix[local_i][local_j] += value * colShapeValues[j];
          }
        }
      }
    }


    template <class Context, class QuadratureRule,
              class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   RowNode const& node,
                                   ColNode const& /*colNode*/,
                                   ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& localFE = node.finiteElement();
      std::size_t size = localFE.size();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      LocalBasisCache<LocalBasisType> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQp(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& shapeValues = shapeValuesCache[iq];

        for (std::size_t i = 0; i < size; ++i) {
          const auto local_i = node.localIndex(i);

          const auto value = factor * shapeValues[i];
          elementMatrix[local_i][local_i] += value * shapeValues[i];

          for (std::size_t j = i+1; j < size; ++j) {
            const auto local_j = node.localIndex(j);

            elementMatrix[local_i][local_j] += value * shapeValues[j];
            elementMatrix[local_j][local_i] += value * shapeValues[j];
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
