#pragma once

#include <cmath>

#include <dune/common/timer.hh>

#include <amdis/Assembler.hpp>
#include <amdis/LocalAssembler.hpp>
#include <amdis/LocalAssemblerList.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/common/Math.hpp>
#include <amdis/utility/MultiIndex.hpp>
#include <amdis/utility/TreePath.hpp>

namespace AMDiS
{
  template <class RowBasisType, class ColBasisType, class Backend>
  class DOFMatrixBase
  {
  public:
    /// The type of the finite element space / basis of the row
    using RowBasis = RowBasisType;
    using RowLocalView = typename RowBasis::LocalView;

    /// The type of the finite element space / basis of the column
    using ColBasis = ColBasisType;
    using ColLocalView = typename ColBasis::LocalView;

    using Element = typename RowLocalView::Element;
    using Geometry = typename Element::Geometry;

    /// The index/size - type
    using size_type = typename RowBasis::size_type;

    using BaseMatrix = typename Backend::BaseMatrix;
    using ElementMatrix = Dune::DynamicMatrix<double>;

  public:
    /// Constructor. Constructs new BaseVector.
    DOFMatrixBase(RowBasis const& rowBasis, ColBasis const& colBasis)
      : rowBasis_(&rowBasis)
      , colBasis_(&colBasis)
      , rowLocalView_(rowBasis.localView())
      , colLocalView_(colBasis.localView())
    {
      operators_.init(rowLocalView_.tree(), colLocalView_.tree(), tag::store{});
    }

    /// Return the row-basis \ref rowBasis of the matrix
    RowBasis const& rowBasis() const
    {
      return *rowBasis_;
    }

    /// Return the col-basis \ref colBasis of the matrix
    ColBasis const& colBasis() const
    {
      return *colBasis_;
    }

    /// Return the data-matrix
    BaseMatrix const& matrix() const
    {
      return backend_.matrix();
    }

    /// Return the data-matrix
    BaseMatrix& matrix()
    {
      return backend_.matrix();
    }

    /// Return the size of the \ref rowBasis_
    size_type rows() const
    {
      return rowBasis_->dimension();
    }

    /// Return the size of the \ref colBasis_
    size_type cols() const
    {
      return colBasis_->dimension();
    }

    /// Initialize the matrix for insertion, e.g. allocate the non-zero pattern
    /// If \p setToZero is true, the matrix is set to 0
    void init(bool asmMatrix)
    {
      backend_.init(*rowBasis_, *colBasis_, asmMatrix);

      auto const rowSize = rowBasis_->localView().maxSize();
      auto const colSize = colBasis_->localView().maxSize();
      elementMatrix_.resize(rowSize, colSize);
    }

    /// Finish the matrix insertion, e.g. cleanup or final insertion
    void finish(bool asmMatrix)
    {
      backend_.finish();

      if (asmMatrix) {
        forEachNode_(rowBasis_->localView().tree(), [&,this](auto const& rowNode, auto&&) {
          forEachNode_(colBasis_->localView().tree(), [&,this](auto const& colNode, auto&&) {
            auto& matOp = operators_[rowNode][colNode];
            if (matOp.doAssemble())
              matOp.assembled = true;
          });
        });
      }
    }

    /// Insert a block of values into the matrix (add to existing values)
    void insert(RowLocalView const& rowLocalView,
                ColLocalView const& colLocalView,
                ElementMatrix const& elementMatrix)
    {
      for (size_type i = 0; i < rowLocalView.size(); ++i) {
        size_type const row = flatMultiIndex(rowLocalView.index(i));
        for (size_type j = 0; j < colLocalView.size(); ++j) {
          if (std::abs(elementMatrix[i][j]) > threshold<double>) {
            size_type const col = flatMultiIndex(colLocalView.index(j));
            backend_.insert(row, col, elementMatrix[i][j]);
          }
        }
      }
    }

    /// Insert a single value into the matrix (add to existing value)
    template <class RowIndex, class ColIndex>
    void insert(RowIndex row, ColIndex col, typename Backend::value_type const& value)
    {
      backend_.insert(flatMultiIndex(row), flatMultiIndex(col), value);
    }

    template <class RowNode, class ColNode>
    auto& operators(RowNode const& row, ColNode const& col)
    {
      return operators_[row][col];
    }

    template <class ContextTag, class Operator,
              class RowTreePath = RootTreePath, class ColTreePath = RootTreePath>
    void addOperator(ContextTag contextTag, Operator const& preOp,
                     RowTreePath row = {}, ColTreePath col = {})
    {
      static_assert( Concepts::PreTreePath<RowTreePath>,
          "row must be a valid treepath, or an integer/index-constant");
      static_assert( Concepts::PreTreePath<ColTreePath>,
          "col must be a valid treepath, or an integer/index-constant");

      auto i = child(rowBasis_->localView().tree(), makeTreePath(row));
      auto j = child(colBasis_->localView().tree(), makeTreePath(col));

      using Context = typename ContextTag::type;
      auto op = makeLocalOperator<Context>(preOp, rowBasis_->gridView());
      auto localAssembler = makeLocalAssemblerPtr<Context>(std::move(op), i, j);

      operators_[i][j].push(contextTag, localAssembler);
      operators_[i][j].changing = true;
    }

    // Assemble the operators on the `element`. Assume that `rowLocalView`
    // and `colLocalView` are already bound to the element.
    void assemble(Element const& element, Geometry const& geometry,
                  RowLocalView const& rowLocalView,
                  ColLocalView const& colLocalView)
    {
      elementMatrix_ = 0;

      forEachNode_(rowLocalView.tree(), [&](auto const& rowNode, auto&&) {
        forEachNode_(colLocalView.tree(), [&](auto const& colNode, auto&&) {
          auto& matOp = operators_[rowNode][colNode];
          if (matOp.doAssemble() && !matOp.empty()) {
            matOp.bind(element, geometry);
            auto matAssembler = [&](auto const& context, auto& operators) {
              for (auto scaled : operators)
                scaled.op->assemble(context, rowNode, colNode, elementMatrix_);
            };

            assembleOperators(rowBasis_->gridView(), element, matOp, matAssembler);
            matOp.unbind();
          }
        });
      });

      insert(rowLocalView, colLocalView, elementMatrix_);
    }

    /// \brief Deletes all rows with \p dirichletNodes != 0 and if \p apply is true, adds
    /// a one on the diagonal of the matrix.
    auto applyDirichletBC(std::vector<char> const& dirichletNodes)
    {
      return backend_.applyDirichletBC(dirichletNodes);
    }

    size_type nnz() const
    {
      return backend_.nnz();
    }

  protected:
    /// The finite element space / basis associated with the data matrix
    RowBasis const*	rowBasis_;
    ColBasis const*	colBasis_;

    /// The LocalViews aassociated with the row/col basis
    RowLocalView rowLocalView_;
    ColLocalView colLocalView_;

    /// Data backend
    Backend backend_;

    /// Dense matrix to store coefficients during \ref assemble()
    ElementMatrix elementMatrix_;

    /// List of operators associated to row/col node
    MatrixOperators<RowBasis,ColBasis> operators_;
  };

} // end namespace AMDiS
